#   Copyright (c) 2020 Boulder Real Time Technologies, Inc.
#       
#   This software module is wholly owned by Boulder Real Time
#   Technologies, Inc. Any use of this software module without
#   express written permission from Boulder Real Time Technologies,
#   Inc. is prohibited.
#

import os
import re
import shutil
import subprocess
from datetime import datetime
import logging

from webservice_fdsn.app.core.waveform_helpers import exceptions

from antelope.datascope import closing
from antelope.datascope import freeing
from antelope.datascope import dbopen
from antelope.datascope import DbsubsetError

log = logging.getLogger(__name__)


def _create_symlink_to_wf(source_wf_path: str, app_path: str) -> None:
    """
    Create symlink to waveform dir in original db path (only if it doesn't exist already).
    This symlink will be used by trexcerpt to look for waveform data
    :param source_wf_path:
    :param app_path:
    :return:
    """
    symlink_path = os.path.join(app_path)
    # First check if link already exists
    link_exists = os.path.lexists(symlink_path)

    if not link_exists:
        log.debug("Creating symlink to source waveform directory...")

        # If link path is a nested directory, create parent directories first and then link to the leaf directory
        if not os.path.isdir(os.path.dirname(symlink_path)):
            os.makedirs(os.path.dirname(symlink_path))
        os.symlink(source_wf_path, symlink_path)
        log.debug(f'Created symlink {symlink_path} -> {source_wf_path}')
    else:
        log.debug(f'Link exists: {symlink_path}')

    # If link exists, check if it is pointing to a valid path
    try:
        os.stat(symlink_path)
    except FileNotFoundError:
        log.error(f'Location trying to symlink to [{source_wf_path}] does not exist, cannot proceed...')
        raise exceptions.InvalidAppConfig(f'!! CRITICAL !!: Antelope waveform data at {source_wf_path} does not '
                                          f'exist. Cannot service dataselect queries.')


def clear_artifacts(paths):
    """
    Clean out intermediate files
    :param list paths:
    :return:
    """
    for path in paths:
        if os.path.exists(path):
            log.debug(f'Clean up: {path}')
            os.remove(path)


def _triage_trexcerpt_error(stderr):
    """Look for error patterns in trexcerpt stderr, raise appropriate exception
    :param bytes stderr: std error output from trexcerpt
    :raises exceptions.RequestedDataNotFoundError
    :raises exceptions.ResponseTooLarge
    :raises Exception
    """
    if re.search(b"won't overwrite existing file", stderr):
        log.debug("Output shows the file over-write error, this is ok...")
        return
    elif re.search(b"no records from wfdisc after time subset", stderr):
        raise exceptions.RequestedDataNotFoundError("No waveform data found in user defined time range")
    elif re.search(b"more than the allowed maximum", stderr):
        log.error("Too much waveform data requested, reduce request window")
        raise exceptions.ResponseTooLarge("Too much waveform data requested, reduce request window")
    elif re.search(b"errors occurred while loading waveforms for", stderr):
        raise Exception(f"Waveform file not found: {stderr}")
    else:
        raise Exception(stderr)


def merge_files(file_paths, dest_file):
    """
    Merge passed in files into 1
    :param list file_paths: files to merge
    :param str dest_file: file to dump merged data into
    :rtype str
    :return: path to merged file
    """
    if len(file_paths) == 1:
        shutil.move(file_paths[0], dest_file)
    else:
        log.debug(f'Merging all seed data into {dest_file}')
        with open(dest_file, "wb") as dest:
            for f in file_paths:
                with open(f, "rb") as src:
                    shutil.copyfileobj(src, dest)
    return dest_file


def dbselect_helper(db, field):
    """
    Simple wrapper to run, parse and return dbselect queries
    :param db:
    :param field:
    :return:
    """
    command = ["dbselect", db, field]
    log.debug("Running dbselect command: {}".format(' '.join(command)))
    _proc = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = _proc.communicate()
    if _proc.returncode != 0:
        log.error(f'Error making dbselect call, stdout: {stdout}, stderr: {stderr}')
    else:
        _out = stdout.decode('utf-8').split()
        if _out and len(_out) > 0:
            log.debug(f'dbselect found {len(_out)} results. 1st result for reference: {_out[0]}')
        else:
            log.debug(f'dbselect returned no results: {_out}')
        return _out


def get_wf_file_paths(wfdisc_path, waveform_dir, check_exists=False):
    """
    Return unique wf file paths by using dir and dfile values for wf files from wf db
    :param wfdisc_path: Path to wfdisc file from which to extract dir and dfile info
    :param waveform_dir: Path to where waveform files exist, with which the above dir/dfile will be appended
    :param check_exists: If true, check if all files exist
    :return: tuple: (list of file paths on local filesystem, unique_parent_dirs)
    """
    log.debug("get_wf_file_path: wfdisc_path: {}".format(wfdisc_path))
    parent_relative_dirs = dbselect_helper(wfdisc_path, "dir")
    file_names = dbselect_helper(wfdisc_path, "dfile")

    _wf_files = zip(parent_relative_dirs, file_names)
    wf_file_full_paths = [os.path.join(waveform_dir, x[0], x[1]) for x in _wf_files]

    if check_exists:
        for _path in wf_file_full_paths:
            try:
                os.stat(_path)
            except FileNotFoundError:
                raise exceptions.RequestedDataNotFoundError(f'Required wf file not found, cannot proceed: {_path}')
            except PermissionError:
                raise exceptions.InvalidAppConfig(f'App encountered permission issue accessing waveform file {_path}, '
                                                  f'cannot proceed')

    # Want to maintain order of files since we are dealing with time-series data, so eliminate duplicates manually.
    # Using Set could potentially change order of files being returned
    unique_files = []
    for f in wf_file_full_paths:
        if f in unique_files:
            continue
        else:
            unique_files.append(f)
    log.debug(f'Found {len(unique_files)} unique wf files')
    return unique_files, set(parent_relative_dirs)


def handle_dataselect_query_via_datascope(db_path, network, station, channel, location, result_db_name):
    """Run datascope methods to extract data matching user query
    :param db_path:
    :param network:
    :param station:
    :param channel:
    :param location:
    :param result_db_name:
    :return: result_db_name: If created, else raises exception
    """
    log.debug(f'Querying using parameters: {network}/{station}/{channel}')

    with closing(dbopen(db_path, 'r')) as db:
        dbtable = db.lookup(table='wfdisc')
        with freeing(dbtable.join('snetsta')) as dbview:
            log.debug(f'Record count after snetsta: {dbview.record_count}')
            try:
                _s = f'snet=="{network}"'
                log.debug(f'Subset query={_s}')
                sbst = dbview.subset(_s)
            except DbsubsetError:
                raise exceptions.RequestedDataNotFoundError("No data found after applying network filter")
            else:
                log.debug(f'Record count after snet: {sbst.record_count}')
                if sbst.record_count <= 0:
                    raise exceptions.RequestedDataNotFoundError("No data found after applying network filter")

            try:
                sbst_stn = sbst.subset(f'sta=~/{station}/')
            except DbsubsetError:
                raise exceptions.RequestedDataNotFoundError("No data found after applying station filter")
            else:
                log.debug(f'Record count after sta filter: {sbst_stn.record_count}')
                if sbst_stn.record_count <= 0:
                    raise exceptions.RequestedDataNotFoundError("No data found after applying station filter")

            # Filter using channel info
            if ',' in channel:
                channel = channel.split(',')
                channel = [ch.strip() for ch in channel]
                channel_subset_cmd = ' || '.join(['chan == "{}"'.format(x) for x in channel])
            else:
                channel_subset_cmd = f'chan=~/{channel}/'

            try:
                sbst_ch = sbst_stn.subset(channel_subset_cmd)
            except DbsubsetError:
                raise exceptions.RequestedDataNotFoundError("No data found after applying channel filter")
            else:
                log.debug(f'Record count after chan filter: {sbst_ch.record_count}')
                if sbst_ch.record_count <= 0:
                    raise exceptions.RequestedDataNotFoundError("No data found after applying channel filter")

            sbst_ch.unjoin(result_db_name, rewrite=True)
            log.debug(f'Wrote result db files generated by datascope operations: {result_db_name}')

    _expected_file = f'{result_db_name}.wfdisc'
    # At this point, check if a wfdisc file exists, which means requested data was found which trexcerpt can work with
    try:
        os.stat(_expected_file)
    except FileNotFoundError:
        _msg = f'Wfdisc file {_expected_file} not generated. This indicates no data was found in db {db} matching user query'
        log.warning(_msg)
        raise exceptions.RequestedDataNotFoundError(_msg)
    else:
        return result_db_name


def handle_dataselect_query(db, network, station, channel, location, starttime, endtime, scratch_dir="/tmp/"):
    """Make antelope db binary calls to satisfy dataselect queries
    :param db: Absolute path to database on disk itself, not the directory of the db.
    :param network: Waveform network
    :param station: Waveform station
    :param channel: Waveform channel
    :param location: Waveform location
    :param datetime.datetime starttime: time from which to grab waveform data
    :param datetime.datetime endtime: time till which to grab waveform data
    :param scratch_dir: directory to create intermediate files in. Will be cleaned up once query is complete
    :rtype str
    :return: path to waveform file
    """
    log.debug(f'Querying using parameters: {network}/{station}/{channel}/"{starttime}"/"{endtime}"')

    # Check if db is valid
    command = ["dbcheck", db]

    try:
        with open(os.devnull, 'w') as FNULL:
            subprocess.check_call(command, stdout=FNULL)
    except subprocess.CalledProcessError as ex:
        log.error(f'Passed in db \"{db}\" reported invalid by dbcheck, error: {ex}')
        raise exceptions.InvalidInput(f'Invalid value passed in. db at {db} is not valid')

    # Temp artifacts get generated here needed to extract the requested wf data. Track and clean up once query is
    # complete
    artifacts = []

    # Get unique identifiers
    query_time = datetime.now()
    unique_id = query_time.microsecond

    _filter_table = f'{network}_{station}_{channel}_{unique_id}'
    _filter_table = _filter_table.replace("*", "_wc_")
    _filter_table = _filter_table.replace(".", "_dot_")

    # Run datascope operations to extract data matching query params
    datascope_result_db_name = handle_dataselect_query_via_datascope(db, network, station,
                                                                     channel, location,
                                                                     os.path.join(scratch_dir, _filter_table))
    
    log.debug(f'datascope has dumped its filtered data to {datascope_result_db_name}')
    artifacts.extend([datascope_result_db_name,
                     f'{datascope_result_db_name}.wfdisc',
                     f'{datascope_result_db_name}.snetsta'])

    # Now that wfdisc has been generated, find out where the actual wf files reside (using the dir and dfile value from
    # wfdisc file), and create a symlink to them in the scratch dir. This symlink path will be passed to trexcerpt
    # which will be looking for the .w files here to extract waveform data from
    wf_files, wf_dirs = get_wf_file_paths(f'{datascope_result_db_name}.wfdisc',
                                          os.path.dirname(db),
                                          check_exists=True)

    log.debug(f'wf_files: {wf_files}')
    log.debug(f'wf_dirs: {wf_dirs}')
    # wf_dirs are the relative paths within the db dir to where wf files exist.
    # Create symlinks to those so trexcerpt has access to waveform files
    for wf_dir in wf_dirs:
        _create_symlink_to_wf(os.path.join(os.path.dirname(db), wf_dir), os.path.join(scratch_dir, wf_dir))

    # Generate dfile path.. cannot exceed 32 chars. Timestamp is 18 chars, so that leaves 14 for sta+chan
    sta_chan = '{}_{}'.format(station, channel)
    if len(sta_chan) > 14:
        sta_chan = sta_chan[:14]

    dfile = "%Y/%j/{sta_chan}.%Y.%j.{hour}.{min}.{sec}".format(sta_chan=sta_chan,
                                                               hour=query_time.strftime("%H"),
                                                               min=query_time.strftime("%M"),
                                                               sec=query_time.strftime("%S"))

    command_trexcerpt = ["trexcerpt",  "-o", "sd", "-m", "time", "-W", datascope_result_db_name, "-w",
                         dfile,
                         f'{datascope_result_db_name}.wfdisc',
                         os.path.join(scratch_dir, f'{_filter_table}_output'),
                         str(datetime.timestamp(starttime.replace(tzinfo=None))),
                         str(datetime.timestamp(endtime.replace(tzinfo=None)))]

    log.debug(f'Found data for network/station/channel combination, now filtering for time ranges using trexcerpt '
              f'command: {command_trexcerpt}')

    # Set cwd to the directory where the waveforms reside, so the related dir+dfile paths contained with wfdisc
    # file will resolve and trexcerpt finds its waveform files
    _proc = subprocess.Popen(command_trexcerpt, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                             cwd=os.path.dirname(db))
    stdout, stderr = _proc.communicate()

    log.debug(f'stdout from trexcerpt: {stdout}')
    if _proc.returncode != 0:
        log.warning(f'Received non-zero return code [{_proc.returncode}] from trexcerpt')
        if stderr:
            try:
                _triage_trexcerpt_error(stderr)
            except:
                clear_artifacts(artifacts)
                raise
        
    # Check if _output file was generated
    try:
        os.stat(os.path.join(scratch_dir, f'{_filter_table}_output.wfdisc'))
    except FileNotFoundError:
        _msg = "Wfdisc file {} not generated. This indicates no data was found in db {} matching user query"\
            .format(os.path.join(scratch_dir, "{}_output.wfdisc".format(_filter_table)), db)
        log.debug(_msg)
        clear_artifacts(artifacts)
        raise exceptions.RequestedDataNotFoundError(_msg)

    artifacts.extend([os.path.join(scratch_dir, f'{_filter_table}_output.wfdisc'),
                      os.path.join(scratch_dir, f'{_filter_table}_output.lastid')])
 
    # Find file path in created db, need to be cleaned up once final data is obtained
    wf_file_paths, _ = get_wf_file_paths(os.path.join(scratch_dir,
                                                      f'{_filter_table}_output.wfdisc'),
                                         scratch_dir)
    artifacts.extend(wf_file_paths)

    # wf_file_paths is a list of paths to all files containing requested mseed data. Get unique ones and combine them
    # so a single mseed file gets returned.
    result_file_path = os.path.join(scratch_dir, _filter_table)
    result_file_path = merge_files(wf_file_paths, result_file_path)
    log.debug(f'Created waveform file/s successfully. {result_file_path}')

    # If file being returned to user is same as one generated, don't clean it up
    if result_file_path in artifacts:
        artifacts = list(set(artifacts) - set([result_file_path]))

    clear_artifacts(artifacts)
    log.debug(f'Returning path: {result_file_path}')
    return result_file_path


if __name__ == "__main__":
    handle_dataselect_query("/Users/rohan_work/rtdemo/db/gsn", "II", "A.*", "BHZ_00", "--",
                            "7/27/19 00:00:00.00000", "7/28/19 2:54:00.00000")
