#   Copyright (c) 2021 Boulder Real Time Technologies, Inc.
#
#   This software module is wholly owned by Boulder Real Time
#   Technologies, Inc. Any use of this software module without
#   express written permission from Boulder Real Time Technologies,
#   Inc. is prohibited.
#
#   Author: Ambli LLC (ambli.ll.company@gmail.com)

import os
import unittest
from datetime import datetime, timedelta
from collections import OrderedDict

from webservice_fdsn.app.utils import pf_helper
from webservice_fdsn.app.utils import exceptions
from webservice_fdsn.app.core.event_exporter.export_events_helpers.exceptions import NoEventDataFound
from webservice_fdsn.app.utils import dbcluster_resolver
from webservice_fdsn.app.utils import utilities
from webservice_fdsn.app.core.constants import service_interfaces_constants
from webservice_fdsn.app.core.constants import dbcluster_constants
from webservice_fdsn.app.core.constants import common


class TestPfHelper(unittest.TestCase):

    def tearDown(self):
        if common.ENV_VAR_PF_PATH in os.environ:
            os.environ.pop(common.ENV_VAR_PF_PATH)

    def test_pf_helper(self):
        """

        :return:
        """
        pf_file_path = os.path.join(os.path.dirname(__file__),
                                    'test_data', 'webservice_fdsn.pf')
        os.environ[common.ENV_VAR_PF_PATH] = pf_file_path
        self.assertIsInstance(pf_helper.get_pf_options(), dict)
        self.assertIsInstance(pf_helper.get_project_scratch_dir(), str)
        self.assertIsInstance(pf_helper.get_default_service_network(), str)
        self.assertIsInstance(pf_helper.get_project_database_path(), str)
        self.assertEqual(pf_helper.get_cluster_name_for_api(service_interfaces_constants.EVENT),
                         'catalog')

    def test_pf_helper_bad_db_path(self):
        """

        :return:
        """ 
        os.environ[common.ENV_VAR_PF_PATH] = os.path.join(os.path.dirname(__file__), os.pardir,
                                                          'test_data', 'webservice_fdsn_missing_db.pf')
        with self.assertRaises(exceptions.RequiredPfOptionMissing):
            pf_helper.get_project_database_path()

        os.environ[common.ENV_VAR_PF_PATH] = os.path.join(os.path.dirname(__file__), os.pardir,
                                                          'test_data', 'webservice_fdsn_invalid_db.pf')
        with self.assertRaises(exceptions.InvalidPfValue):
            pf_helper.get_project_database_path()


class TestDBCentralResolver(unittest.TestCase):

    def setUp(self):
        self.dbcentral_path = os.path.join(os.path.dirname(__file__),
                                           os.pardir, 'test_data',
                                           'dbcentral7v1')
        self.dbc = dbcluster_resolver.DBClusterResolver(self.dbcentral_path)
    
    def test_happy_path(self):
        self.assertEqual(len(self.dbc.clusters), 3)
        self.assertIn('waveforms', self.dbc.clusters)
        self.assertIn('catalog', self.dbc.clusters)
        self.assertIn('master_stations', self.dbc.clusters)

    def test_stations_dfiles(self):
        dfiles = self.dbc.get_dfiles_by_date(service_interfaces_constants.STATION,
                                             datetime.now(),
                                             datetime.now() + timedelta(days=2))
        # stations has a single file
        self.assertEqual(len(dfiles), 1)

    def test_catalog_dfiles(self):
        dfiles = self.dbc.get_dfiles_by_date(service_interfaces_constants.EVENT,
                                             datetime(year=2020, day=1, month=10),
                                             datetime(year=2021, day=5, month=1))
        self.assertEqual(len(dfiles), 4)

        dfiles = self.dbc.get_dfiles_by_date(service_interfaces_constants.EVENT,
                                             datetime(year=2020, day=1, month=10),
                                             datetime(year=2020, day=19, month=10))
        self.assertEqual(len(dfiles), 1)

        dfiles = self.dbc.get_dfiles_by_date(service_interfaces_constants.EVENT,
                                             datetime(year=2020, day=1, month=10),
                                             datetime(year=2020, day=5, month=11))
        self.assertEqual(len(dfiles), 2)

        dfiles = self.dbc.get_dfiles_by_date(service_interfaces_constants.EVENT,
                                             datetime(year=2020, day=1, month=10),
                                             datetime(year=2021, day=5, month=3))
        self.assertEqual(len(dfiles), 6)

        dfiles = self.dbc.get_dfiles_by_date(service_interfaces_constants.EVENT,
                                             datetime(year=2020, day=1, month=10),
                                             datetime(year=2021, day=5, month=3), sanitize=True)
        self.assertEqual(len(dfiles), 2)

    def test_waveform_dfiles(self):
        dfiles = self.dbc.get_dfiles_by_date(service_interfaces_constants.DATASELECT,
                                             datetime(year=2020, day=1, month=1),
                                             datetime(year=2020, day=19, month=1),
                                             sanitize=False)
        self.assertEqual(len(dfiles), 19)
        dfiles = self.dbc.get_dfiles_by_date(service_interfaces_constants.DATASELECT,
                                             datetime(year=2020, day=1, month=1),
                                             datetime(year=2020, day=19, month=1),
                                             sanitize=True)
        self.assertEqual(len(dfiles), 10)

        dfiles = self.dbc.get_dfiles_by_date(service_interfaces_constants.DATASELECT,
                                             datetime(year=1991, day=1, month=1),
                                             datetime(year=1991, day=19, month=1),
                                             sanitize=True)
        self.assertEqual(len(dfiles), 0)

        dfiles = self.dbc.get_dfiles_by_date(service_interfaces_constants.DATASELECT,
                                             datetime(year=2018, day=1, month=1,
                                                      hour=22, minute=0, second=0),
                                             datetime(year=2018, day=2, month=1,
                                                      hour=4, minute=0, second=0),
                                             sanitize=False)
        self.assertEqual(len(dfiles), 2)

        dfiles = self.dbc.get_dfiles_by_date(service_interfaces_constants.DATASELECT,
                                             datetime(year=2018, day=1, month=1,
                                                      hour=22, minute=0, second=0),
                                             datetime(year=2018, day=2, month=1,
                                                      hour=23, minute=59, second=59),
                                             sanitize=False)
        self.assertEqual(len(dfiles), 2)
        # Check if the correct dates are returned too, not just the number
        expected = set(['waveforms_2018_01_01', 'waveforms_2018_01_02'])
        returned = set([os.path.basename(p) for p in dfiles])
        self.assertEqual(expected, returned)

        dfiles = self.dbc.get_dfiles_by_date(service_interfaces_constants.DATASELECT,
                                             datetime(year=2018, day=1, month=1,
                                                      hour=22, minute=0, second=0),
                                             datetime(year=2018, day=3, month=1,
                                                      hour=0, minute=0, second=0),
                                             sanitize=False)
        self.assertEqual(len(dfiles), 3)
        # Check if the correct dates are returned too, not just the number
        expected = set(['waveforms_2018_01_01', 'waveforms_2018_01_02', 'waveforms_2018_01_03'])
        returned = set([os.path.basename(p) for p in dfiles])
        self.assertEqual(expected, returned)

        dfiles = self.dbc.get_dfiles_by_date(service_interfaces_constants.DATASELECT,
                                             datetime(year=2018, day=1, month=1,
                                                      hour=4, minute=0, second=0),
                                             datetime(year=2018, day=1, month=1,
                                                      hour=14, minute=0, second=0),
                                             sanitize=False)
        self.assertEqual(len(dfiles), 1)
        self.assertIn('2018_01_01', dfiles[0])

    def test_invalid_enddate(self):
        with self.assertRaises(Exception):
            self.dbc.get_dfiles_by_date(service_interfaces_constants.DATASELECT,
                                        datetime(year=2020, day=2, month=1),
                                        datetime(year=2020, day=1, month=1))

    def test_invalid_clustername(self):
        with self.assertRaises(exceptions.RequiredPfOptionMissing):
            self.dbc.get_dfiles_by_date('invalid', datetime(year=2020, day=2, month=1),
                                        datetime(year=2020, day=1, month=1))

    def test_invalid_dbcluster_path(self):
        with self.assertRaises(Exception):
            dbcluster_resolver.DBClusterResolver('/tmp/somenonexistantfile')


class TestUtilities(unittest.TestCase):

    def test_handle_limit_and_offset_on_list(self):
        data = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

        # Test handling of evids of type list
        self.assertEqual([1], utilities.handle_limit_and_offset(data, limit=1, offset=0))
        self.assertEqual([2], utilities.handle_limit_and_offset(data, limit=1, offset=1))
        self.assertEqual([1, 2, 3], utilities.handle_limit_and_offset(data, limit=3, offset=0))
        self.assertEqual([6, 7, 8, 9, 10], utilities.handle_limit_and_offset(data, limit=7, offset=5))
        self.assertEqual([6, 7], utilities.handle_limit_and_offset(data, limit=2, offset=5))
        with self.assertRaises(NoEventDataFound):
            self.assertEqual([], utilities.handle_limit_and_offset(data, limit=1, offset=11))
        self.assertEqual([1], utilities.handle_limit_and_offset(data, limit=1, offset=0))
        with self.assertRaises(NoEventDataFound):
            utilities.handle_limit_and_offset(data, limit=0, offset=0)

        data = [OrderedDict({'a': 'b', 'c': 'd'})]
        self.assertEqual(data, utilities.handle_limit_and_offset(data, limit=2, offset=0))
        with self.assertRaises(NoEventDataFound):
            utilities.handle_limit_and_offset(data, limit=1, offset=1)
        with self.assertRaises(NoEventDataFound):
            utilities.handle_limit_and_offset(data, limit=0, offset=0)
        with self.assertRaises(Exception):
            utilities.handle_limit_and_offset(1, limit=1, offset=0)


    def test_combine_quakemls(self):
        """Exercise combining quakemls"""
        with open(os.path.join(os.path.dirname(__file__),
                               os.pardir, 'test_data',
                               'test_quakeml.xml')) as f:
            quakeml = f.read()

        result = utilities.combine_quakemls([quakeml, quakeml])
        self.assertIsInstance(result, str)
        # Check returned length is greater than single quakeml
        self.assertGreater(len(result), len(quakeml))

    def test_combine_quakemls_limit_offset(self):
        """Exercise combining quakemls"""
        with open(os.path.join(os.path.dirname(__file__),
                               os.pardir, 'test_data',
                               'test_quakeml.xml')) as f:
            quakeml = f.read()

        result = utilities.combine_quakemls([quakeml, quakeml], limit=1, offset=0)
        self.assertIsInstance(result, str)
        result = utilities.combine_quakemls([quakeml, quakeml], limit=4, offset=2)
        self.assertIsInstance(result, str)
        result = utilities.combine_quakemls([quakeml, quakeml], offset=3)
        self.assertIsInstance(result, str)

    def test_get_db_schema(self):
        """Test fetching db schema"""
        dbcluster_path = os.path.join(os.path.dirname(__file__),
                                      os.pardir, 'test_data',
                                     'dbcentral7v1')
        self.assertEqual(utilities.get_db_schema(dbcluster_path), dbcluster_constants.DBCLUSTER07)

        # This path always exists inside antelope installations
        dbcluster_path = '/opt/antelope/data/db/demo/demo'
        self.assertEqual(utilities.get_db_schema(dbcluster_path), dbcluster_constants.CSS3_0)
